/*	fsacl.cc		*/

/* Copyright (C) Jan Daciuk, 2016 */

#include	<string>
#include	<iostream>
#include	<fstream>
#include	<algorithm>
#include	<vector>
#include	<set>
#include	<map>
#include	<iterator>
#include	<string.h>
#include	"fsacl.h"




class ErrorException {
public:
  int	error_code;
  ErrorException(const int c) : error_code(c) {}
};

static int fsacl_errno;

const unsigned int REGISTER_SIZE = 9001;

class state;


struct bbrother_struct {
  stateno_t	state_no;      // small brother
  stateno_t	bb;		// big brother state number
  int		offset;		// how many initial transitions to skip
  bbrother_struct(const stateno_t s, const stateno_t b, const int o)
    : state_no(s), bb(b), offset(o) {}
};

struct signature {              /* dictionary file signature */
  char          sig[4];         /* automaton identifier (magic number) */
  char          ver;            /* automaton type number */
  char          filler;         /* char used as filler */
  char          annot_sep;      /* char that separates annotations from lex */
  char          gtl;            /* length of go_to field (gtl) */
                                /* when compiled with NUMBERS, the upper */
                                /* 4 bits contain ctl - length of a field */
                                /* before transitions that hold the number */
                                /* of strings */
                                /* recognized by the automaton starting */
                                /* from given node */
};/*struct signature */

class transition {
protected:
  stateno_t		next;		// next transition of the same state
  					// or the next empty transition
  stateno_t		target;		// target state of the transition
  unsigned char		label;		// label of the transition
  unsigned char		flags;		// finality, etc.
  // Flags in flags field:
  // bitmask	name	used for
  // 1		final	marking final transitions
  // 2		next	marking transitions with the NEXT flag
public:
  // Create a new transition (used internally)
  transition(void) : next(-1), target(0), label(0), flags(0) {}
  // Create a copy of another transition
  transition(const stateno_t t, const unsigned char l)
    : next(-1), target(t), label(l), flags(0) {}
  // Create a copy of another transition
  transition(const stateno_t n, const stateno_t t, const unsigned char l,
	     const int f)
    : next(n), target(t), label(l), flags(f) {}
  // Set the index of the next transition
  void set_next(const stateno_t n) { next = n; }
  // get the index of the next transition
  stateno_t get_next(void) const { return next; }
  void set_target(const stateno_t t) { target = t; }
  stateno_t get_target(void) const { return target; }
  void set_label(const unsigned char l) { label = l; }
  unsigned char get_label(void) const { return label; }
  void set_final(const int f) { if (f) flags |= 1; else flags &= ~1; }
  int get_final(void) const { return flags & 1; }
  void set_next_flag(const int n) { if (n) flags |= 2; else flags &= ~2; }
  int get_next_flag(void) const { return flags & 2; }
}; // transition

class state {
protected:
  stateno_t	first_transition; // first transition number
  unsigned int	hit_count;	// number of incoming transitions
  unsigned int	ntrans;		// number of outgoing transitions
  unsigned int	flags;		// various markers
  // Flags in flags field:
  // bitmask	name	used for
  // 1		written	marking already written states
  // 2		deleted	marking deleted states
  // 4		bb	marking states that are big brothers of other states
  // 8		sb	marking states that are smaller brothers of other
  // 16	   	mark1	marking already numbered entries
  // 32		cnxt	marking states already examined for the next flag
  // 64		addrcmpt	marking states that already have an adress
  // 128	reduced marking states that have been visited in reduction
  // 254	redundant marking states that have an equivalent
  // 512	waste   marking states that are no longer needed
  // 1024	next	marking states that have a NEXT flag on last transition
  static stateno_t	total_states; // for statistics
  static stateno_t	max_states;   // for statistics
public:
  state(void) : first_transition(-1), hit_count(0), ntrans(0), flags(0) {}
  state(const state& s)
    : first_transition(s.first_transition), hit_count(s.hit_count), ntrans(s.ntrans),
      flags(0) {}
  void reset(void) {
    first_transition = -1; hit_count = 0; ntrans = 0; flags = 0;
  }
  void set_transitions(const stateno_t t) { first_transition = t; }
  stateno_t get_transitions(void) const { return first_transition; }
  int get_hit_count(void) const { return hit_count; }
  unsigned int hit(void) { return ++hit_count; }
  unsigned int unhit(void) { return --hit_count; }
  unsigned int get_ntrans(void) const { return ntrans; }
  void set_ntrans(const unsigned int n) { ntrans = n; }
  void inc_ntrans(void) { ntrans++; }
  void set_written(void) {flags |= 1; }
  bool is_written(void) { return flags & 1; }
  void set_deleted(void) { flags |= 2; }
  void clear_deleted(void) { flags &= ~2; }
  bool is_deleted(void) const { return flags & 2; }
  void set_big_brother_mark(void) { flags |= 4; }
  void clear_big_brother_mark(void) { flags &= ~4; }
  bool is_big_brother(void) const { return flags & 4; }
  void set_small_brother_mark(void) { flags |= 8; }
  void clear_small_brother_mark(void) { flags &= ~8; }
  bool is_small_brother(void) const { return flags & 8; }
  void set_mark1(void) { flags |= 16; }
  void clear_mark1(void) { flags &= ~16; }
  bool is_mark1(void) const { return flags & 16; }
  void set_cnxt(void) { flags |= 32; }
  void clear_cnxt(void) { flags &= ~32; }
  bool is_cnxt(void) const { return flags & 32; }
  bool can_be_next(void) const {
    // A transition with this state as target can hold the NEXT flag
    return (ntrans > 0 && !is_cnxt() && !is_small_brother());
  }
  void set_addrcmpt(void) { flags |= 64; }
  void clear_addrcmpt(void) { flags &= ~64; }
  bool is_addrcmpt(void) const { return flags & 64; }
  bool is_reduced(void) const {return flags & 128; }
  void set_reduced(void) { flags |= 128; }
  void clear_reduced(void) { flags &= ~128; }
  bool is_redundant(void) const { return flags & 256; }
  void set_redundant(void) { flags |= 256; }
  void clear_redundant(void) { flags &= ~256; }
  bool is_waste(void) const { return flags & 512; }
  void set_waste(void) { flags |= 512; }
  void clear_waste(void) { flags &= ~512; }
  bool is_processed(void) const { return flags & (128+256+512); }
  bool has_next_flag(void) const { return flags & 1024; }
  void set_next_flag(void) { flags |= 1024; }
  void clear_next_flag(void) { flags &= ~1024; }
}; // state

stateno_t state::total_states = 0;
stateno_t state::max_states = 0;

class hash_pointer;

class automaton {
protected:
  // The first in a chain of gaps in place of deleted states
  stateno_t			deleted_state;
  // The first in a chain of gaps in place of deleted transitions
  stateno_t			deleted_transition;
  unsigned int			options; // options (bit masks) for building
  stateno_t			number_of_states;
  stateno_t			number_of_transitions;
  stateno_t			*entries; // cardinalities of right language
  stateno_t			root, superroot; // for compatibility
  stateaddr_t			current_loc;	 // addr of next location
  unsigned int			gtl;		 // length of pointers
  std::vector<state>		states;
  std::vector<transition>	transitions;
  /*std::vector<std::set<stateno_t> >	Register;*/
  std::vector<bbrother_struct>	bbrother;
  stateaddr_t			*stateaddr;
  unsigned long int		number_of_bbrothers;
  unsigned long int		entryl;
  char				annot_separator;
  char				filler;
public:
  static const int			flags_size = 3; // FINAL,STOP,NEXT
  int cmp_states(const stateno_t s1, const stateno_t s2) const {
#ifdef DEBUG
    std::cerr << "cmp_states(" << s1 << ", " << s2 << ")" << std::endl
	      << "ntrans:" << states[s1].get_ntrans() << " and "
	      << states[s2].get_ntrans() << std::endl;
#endif
    int c = (states[s1].get_ntrans() - states[s2].get_ntrans());
    if (c != 0) {
      return c;
    }
    stateno_t t1 = states[s1].get_transitions();
    stateno_t t2 = states[s2].get_transitions();
    while (t1 != -1 && t2 != -1) {
      c = (transitions[t1].get_label() - transitions[t2].get_label());
      if (c != 0) {
	return c;
      }
      else {
	c = (transitions[t1].get_target() - transitions[t2].get_target());
	if (c != 0) {
	  return c;
	}
	else {
	  c = (transitions[t1].get_final() - transitions[t2].get_final());
	  if (c != 0) {
	    return c;
	  }
	}
      }
      t1 = transitions[t1].get_next(); t2 = transitions[t2].get_next();
    }
    if (t1 == t2) {
      return 0;
    }
    else if (t1 < t2) {
      return -1;
    }
    else {
      return 1;
    }
  }
  automaton(void)
    : deleted_state(-1), deleted_transition(-1), number_of_states(0),
      number_of_transitions(0), entries(NULL) {
    options = 0;
    stateno_t null_state = new_state(); // should be 0
    stateno_t null_transition = add_transition(null_state, 0, 0, 0);
    superroot = new_state(); // for compability
    root = new_state();
    add_transition(superroot, root, '^', 0); // for compatibility
    stateaddr = 0;
    filler = '_'; annot_separator = '+';
    number_of_bbrothers = 0;
  }
  ~automaton(void) {
    delete [] entries;
    delete [] stateaddr;
  }
  std::vector<state> *get_states(void) { return &states; }
  std::vector<transition> *get_transitions(void) { return &transitions; }
  state get_state(const stateno_t state_no) const { return states[state_no]; }
  stateno_t new_state(void);
  stateno_t new_state(const stateno_t state_no);
  void delete_state(const stateno_t state_no);
  void delete_useless_state(const stateno_t state_no) {
    states[state_no].unhit();
    if (states[state_no].get_hit_count() < 1) {
      delete_state(state_no);
    }
  }
  stateno_t new_transition(const stateno_t next, const stateno_t target,
			   const unsigned char label, const int final);
  stateno_t add_transition(const stateno_t source, const stateno_t target,
		     const unsigned char label, const int final = 0);
  void add_transitions(const stateno_t state_no, const stateno_t trans_no);
  void change_transition(const stateno_t source, const stateno_t target,
			 const unsigned char label);
  void change_transition(const stateno_t source, const stateno_t target,
			 const unsigned char label, int fin);
  void delete_transition(const stateno_t transition_no);
  stateno_t delta(const stateno_t source, const unsigned char label);
  stateno_t register_find_or_put(const stateno_t state_no);
  void remove_from_register(const stateno_t s);
  unsigned long int hash(const stateno_t state_no);
  stateno_t find_equiv(const stateno_t state_no, const unsigned int skip);
  stateno_t get_root(void) const { return root; }
  stateno_t get_superroot(void) const { return superroot; }
  long int number_entries(const stateno_t state_no);
  void share_arcs(void);
  stateno_t count_arcs(void);
  stateno_t count_next(const stateno_t state_no);
  stateaddr_t compute_addresses(const stateno_t state_no);
  stateno_t reduce_inner(const stateno_t state_no);
  void write_automaton(std::ostream &output);
  void write_header(std::ostream &output);
  void write_transition_vector(std::ostream &output);
  void write_state(std::ostream &output, const stateno_t state_no);
  stateaddr_t serialize_transition(char *buffer, const transition &t,
				   const unsigned int start);
  void write_any_construction_file(std::ostream & output);
  void set_option(const unsigned int opt) { options |= opt; }
  void clear_option(const unsigned int opt) { options &= ~opt; }
  unsigned int get_options(void) const { return options; }
  void set_annot_separator(const char s) { annot_separator = s; }
  void set_filler(const char f) { filler = f; }
  void sort_transitions(void);
}; //automaton

struct less_states {
  static automaton *a;
  void set_automaton (automaton *aut) { a = aut; }
  bool operator()(const stateno_t state1, const stateno_t state2) const {
    return a->cmp_states(state1, state2) < 0;
  }
};

std::vector<std::set<stateno_t, less_states> > Register;


/* The only purpose of class hash_pointer is to get access to transitions
 * of states given their state numbers in comparison functions
 * used in the register for comparison inside std::set
 */
class hash_pointer {
protected:
  stateno_t		p;	// state number
  static automaton	*a;	// access to methods of states and transitions
public:
  // turn state number into a state number with access to methods of automaton
  hash_pointer(const stateno_t s) : p(s) {}
  // provide a pointer to automaton so that methods of automaton are accessible
  hash_pointer(automaton *aut) : p(0)  { a = aut; }
  bool operator==(hash_pointer &s) const {
    return (a->cmp_states(p, s.p) == 0);
  }
  bool operator <(const hash_pointer &s) const {
    return (a->cmp_states(p, s.p) < 0);
  }
}; // hash_pointer


/* Name:	initialize_Register
 * Class:	None
 * Purpose:	Fills the Register with empty sets for each possible hash value.
 * Parameters:	a		- (i) automaton, for the Register is used;
 *		l		- (o) less_states structure that needs
 *					to be initialized with the automaton.
 * Returns:	Nothing.
 * Globals:	Register	- (o) set of states with already fixed right
 *					language.
 * Exceptions:	None.
 * Remarks:	The Register is implemented as a vector of sets of state
 *		numbers, with comparison of state numbers based on finality
 *		and the set of outgoing transitions of states.
 *		The Register is a structure that gave me the biggest headache
 *		in the development of the library. It apparently must be
 *		a global variable. It cannot be a field of the automaton class,
 *		because every way I tried it, it created circular definictions.
 *
 *		The Register is stored in fsacl.cc.
 */
void
initialize_Register(automaton &a, less_states &l);


/* Name:	initialize_Register
 * Class:	None
 * Purpose:	Fills the Register with empty sets for each possible hash value.
 * Parameters:	a		- (i) automaton, for the Register is used;
 *		l		- (o) less_states structure that needs
 *					to be initialized with the automaton.
 * Returns:	Nothing.
 * Globals:	Register	- (o) set of states with already fixed right
 *					language;
 *		REGISTER_SIZE	- (i) size of the Register.
 * Exceptions:	None.
 * Remarks:	The Register is implemented as a vector of sets of state
 *		numbers, with comparison of state numbers based on finality
 *		and the set of outgoing transitions of states.
 *		The Register is a structure that gave me the biggest headache
 *		in the development of the library. It apparently must be
 *		a global variable. It cannot be a field of the automaton class,
 *		because every way I tried it, it created circular definictions.
 *
 *		The Register is stored in fsacl.cc.
 */
void initialize_Register(automaton &a, less_states &l) {
  l.set_automaton(&a);
  std::set<stateno_t,less_states> ss;
  Register = std::vector<std::set<stateno_t,less_states> >(REGISTER_SIZE, ss);
}//initialize_Register


automaton *less_states::a = 0;


/* Name:	bits_in
 * Class:	None.
 * Purpose:	Calculates the length of an unsigned number in bits.
 * Parameters:	n		- (i) the number to be examined.
 * Returns:	The minimum number of bits needed to store number n.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	It takes 1 bit to store 0.
 */
unsigned int
bits_in(unsigned long int n) {
  unsigned int bits = 0;
  if (n <= 0) bits = 1;
  while (n > 0) {
    n >>= 1; bits++;
  }
  return bits;
}//bits_in

/* Name:	new_state
 * Class:	automaton
 * Purpose:	Creates a new state.
 * Parameters:	None.
 * Returns:	State number.
 * Globals:	None.
 * Exceptions:	Standard exceptions of std::vector.
 * Remarks:	The new state has no transitions.
 *		States are stored in a vector "states".
 *		Deletion of states inside that vector creates holes
 *		that can be used as places for new states. Deleted
 *		states are organized in a list that starts with
 *		a state pointed to by variable "deleted_state".
 *		The field "first_transition" in the list points
 *		to the next state on the list.
 */
stateno_t
automaton::new_state(void) {
  number_of_states++;
  if (deleted_state >= 0) {
    // There are gaps in the states vector
    stateno_t n = deleted_state;
    deleted_state = states[deleted_state].get_transitions();
    states[n].reset();
    return n;
  }
  else {
    state s;
    states.push_back(s);
    stateno_t n = states.size() - 1;
    return n;
  }
}//automaton::new_state

/* Name:	new_transition
 * Class:	automaton
 * Purpose:	Creates a new transition.
 * Parameters:	next		- (i) next transition in a list of outgoing
 *					transitions of a state;
 *		target		- (i) target of the transition;
 *		label		- (i) label of the transition;
 *		final		- (i) 0 if the transition is not final.
 * Returns:	The index of the transition in the transitions vector.
 * Globals:	None.
 * Exceptions:	Standard std::vector exceptions.
 * Remarks:	If all slots in the transitions vector are occupied,
 *		the transition is put at the end of the vector, and the size
 *		of the vector is increased.
 *		Deleting a transition can create a hole in the transitions
 *		vector. Such holes are chained using field "next", and
 *		field "deleted_transition" of automaton shows the first deleted
 *		transition on the list. When new_transition is called,
 *		and deleted_transition is not -1, then the first deleted
 *		transition on the list is reused.
 */
stateno_t
automaton::new_transition(const stateno_t next, const stateno_t target,
			  const unsigned char label, const int final) {
  number_of_transitions++;
  if (deleted_transition == -1) {
    // transitions vector has no holes
    transitions.push_back(transition(next, target, label, final));
    return transitions.size() - 1;
  }
  else {
    // Fill in a hole
    stateno_t t = deleted_transition;
    deleted_transition = transitions[t].get_next();
    transitions[t].set_next(next);
    transitions[t].set_target(target);
    transitions[t].set_label(label);
    transitions[t].set_final(final);
    return t;
  }
}//automaton::new_transition

/* Name:	delete_transition
 * Class:	automaton
 * Purpose:	Deletes a transition.
 * Parameters:	transition_no	- (i) transition number.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:	invalid_trans_no when invalid transition number is specified.
 *		Standard std::vector exceptions.
 * Remarks:	If the transition is the last in the transitions vector,
 *		the size of the vector is decreased. Otherwise the transition
 *		is put onto a list of deleted transitions.
 */
void
automaton::delete_transition(const stateno_t trans_no) {
#ifdef DEBUG
  if (trans_no < 0 || trans_no >= transitions.size()) {
    throw ErrorException(FSACL_InvalidTransNo);
  }
#endif
  if (trans_no == transitions.size() - 1) {
    transitions.pop_back();
  }
  else {
    transitions[trans_no].set_next(deleted_transition);
    deleted_transition = trans_no;
  }
  --number_of_transitions;
}//automaton::delete_transition

/* Name:	add_transitions
 * Class:	automaton
 * Parameters:	state_no	- (i) source state for new transitions;
 *		trans_no	- (i) transition number that begins a sequence.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	Since transitions are sorted, it makes sense to add
 *		transitions of one state in reverse order to another state.
 */
void
automaton::add_transitions(const stateno_t state_no, const stateno_t trans_no) {
  if (trans_no != -1) {
    add_transitions(state_no, transitions[trans_no].get_next());
    add_transition(state_no, transitions[trans_no].get_target(),
		   transitions[trans_no].get_label(),
		   transitions[trans_no].get_final());
  }
}//

/* Name:	new_state
 * Class:	automaton
 * Purpose:	Clones an existing state.
 * Parameters:	state_no	- (i) state number to be cloned.
 * Returns:	New state number.
 * Globals:	None.
 * Exceptions:	Standard exceptions of std::vector.
 * Remarks:	The new state is identical to the one that is to be cloned.
 *		However, its transitions have to be copied. It is not
 *		possible to copy only the first transition number.
 */
stateno_t
automaton::new_state(const stateno_t state_no) {
  stateno_t n = new_state();
  add_transitions(n, states[state_no].get_transitions());
}//automaton::new_state

/* Name:	delete_state
 * Class:	automaton
 * Purpose:	Deletes a state with given number.
 * Parameters:	state_no	- (i) state number to be deleted.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:	FSACL_IllegalState - when non-existing state number given.
 * Remarks:	States are stored in a vector "states".
 *		Deletion of states inside that vector creates holes
 *		that can be used as places for new states. Deleted
 *		states are organized in a list that starts with
 *		a state pointed to by variable "deleted_state".
 *		The field "first_transition" in the list points
 *		to the next state on the list.
 */
void
automaton::delete_state(const stateno_t state_no) {
#ifdef DEBUG
  if (state_no >= states.size() || state_no < 0) {
    // No such state
    throw ErrorException(FSACL_IllegalState);
  }
#endif
  // Deal with targets of outgoing transitions, delete transitions
  stateno_t tt = -1;
  for (stateno_t t = states[state_no].get_transitions(); t != -1;
       t = tt) {
    if (states[transitions[t].get_target()].unhit() <= 0) {
      delete_useless_state(transitions[t].get_target()); // deal with target
      tt = transitions[t].get_next();
      delete_transition(t);	// delete the transition
    }
  }
  if (state_no == states.size()) {
    // The last state in the vector - shorten the vector
    states.pop_back();
  }
  else {
    states[state_no].set_transitions(deleted_state);
    states[state_no].set_deleted();
    deleted_state = state_no;
  }
  --number_of_states;
}//automaton::delete_state

/* Name:	add_transition
 * Class:	automaton
 * Purpose:	Adds a new outgoing transition.
 * Parameters:	source		- (i) source transition number;
 *		target		- (i) target transition number;
 *		label		- (i) transition label;
 *		final		- (i) true if transition is final.
 * Returns:	The new transition number.
 * Globals:	None.
 * Exceptions:
 * Remarks:	Outcoming transitions are sorted on labels.
 *		Outcoming transitions of a state are organized in a list.
 *		Adding a transition increases the incoming transition
 *		counter for the target.
 */
stateno_t
automaton::add_transition(const stateno_t source, const stateno_t target,
			  const unsigned char label, const int final) {
#ifdef DEBUG
  std::cerr << "add_transition(" << source << ", " << target << ", "
	    << label << ", " << final << ")" << std::endl;
#endif
  stateno_t t = states[source].get_transitions();
  stateno_t tt = t;		// previous transition
  states[target].hit();
  states[source].inc_ntrans();
  //number_of_transitions++;
  if (t == -1) {
    // First transition for the state
    stateno_t t1 = new_transition(-1, target, label, final);
    states[source].set_transitions(t1);
#ifdef DEBUG
    std::cerr << "Created first transition " << t1 << "(" << target << ", '"
	      << label << "', " << final << ", " << "-1)" << std::endl;
#endif    
    return t1;
  }
  else if (transitions[t].get_label() > label) {
    // The transition is to be put at the start of the list
    stateno_t t1 = new_transition(t, target, label, final);
    states[source].set_transitions(t1);
#ifdef DEBUG
    std::cerr << "Put transition " << t1 << "(" << target << ", '" << label
	      << "', " << final << ", " << t << ") in front of " << t
	      << std::endl;
#endif
    return t1;
  }
  else {
    // The transition is to be put later onto the list of outgoing transitions
    while (t != -1 && transitions[t].get_label() < label) {
#ifdef DEBUG
      std::cerr << "(" << transitions[t].get_target() << ", '"
		<< transitions[t].get_label() << "', "
		<< transitions[t].get_final() << ", "
		<< transitions[t].get_next() << ") ";
#endif      
      tt = t; t = transitions[t].get_next();
    }
    stateno_t t1 = new_transition(t, target, label, final);
#ifdef DEBUG
    std::cerr << "Put new transtion " << t1 << "(" << target << ", '" << label
	      << "', " << final << ", " << t << ")" << std::endl;
#endif    
    transitions[tt].set_next(t1);
    return t1;
  }
}//automaton::add_transition

/* Name:	change_transition
 * Class:	automaton
 * Purpose:	Changes the target of a transition.
 * Parameters:	source		- (i) source state;
 *		target		- (i) target state;
 *		label		- (i) label of the transition.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:	Standard std::vector exceptions.
 * Remarks:	Outcoming transitions are sorted on labels.
 *		Outcoming transitions of a state are organized in a list.
 *		Changing a transition decreases the incoming transition
 *		counter for the previous target, and increases the counter
 *		for the new target.
 *		We assume that the transitions are sorted on labels,
 *		and that a transition of label 'label' goes out from state
 *		'source'.
 */
void
automaton::change_transition(const stateno_t source, const stateno_t target,
			     const unsigned char label) {
  stateno_t t = states[source].get_transitions();
  while (t != -1 && transitions[t].get_label() < label) {
    t = transitions[t].get_next();
  }
  if (transitions[t].get_target() != target) {
    states[target].hit();
    delete_useless_state(transitions[t].get_target());
    transitions[t].set_target(target);
  }
}//automaton::change_transition

/* Name:	change_transition
 * Class:	automaton
 * Purpose:	Changes the target and finality of a transition.
 * Parameters:	source		- (i) source state;
 *		target		- (i) target state;
 *		label		- (i) label of the transition;
 *		fin		- (i) finality of the transition
 *					after the change.
 * Returns:	Transition number.
 * Globals:	None.
 * Exceptions:	Standard std::vector exceptions.
 * Remarks:	Outcoming transitions are sorted on labels.
 *		Outcoming transitions of a state are organized in a list.
 *		Changing a transition decreases the incoming transition
 *		counter for the previous target, and increases the counter
 *		for the new target.
 */
void
automaton::change_transition(const stateno_t source, const stateno_t target,
			     const unsigned char label, const int fin) {
#ifdef DEBUG
  std::cerr << "change_transition(" << source << ", " << target << ", '"
	    << label << "', " << fin << ")" << std::endl;
#endif  
  stateno_t t = states[source].get_transitions();
  while (t != -1 && transitions[t].get_label() < label) {
    t = transitions[t].get_target();
  }
  if (transitions[t].get_target() != target) {
    states[target].hit();
    delete_useless_state(transitions[t].get_target());
    transitions[t].set_target(target);
    transitions[t].set_final(fin);
  }
}//automaton::change_transition

/* Name:	delta
 * Class:	automaton
 * Purpose:	Traverses a transition from the source state on label
 *		giving the target state.
 * Parameters:	source		- (i) source state;
 *		label		- (i) transition label.
 * Returns:	The target state or -1 if transition does not exist.
 * Globals:	None.
 * Exceptions:	Standard exceptions of std::vector.
 * Remarks:	Outgoing transitions of a state are sorted.
 */
stateno_t
automaton::delta(const stateno_t source, const unsigned char label) {
#ifdef DEBUG
  std::cerr << "delta(" << source << ", '" << label << "') -> ";
#endif  
  stateno_t t = states[source].get_transitions();
  while (t != -1 && transitions[t].get_label() <= label) {
    if (transitions[t].get_label() == label) {
#ifdef DEBUG      
      std::cerr << transitions[t].get_target() << std::endl;
#endif      
      return transitions[t].get_target();
    }
    t = transitions[t].get_next();
  }
#ifdef DEBUG  
  std::cerr << -1 << std::endl;
#endif  
  return -1;
}//automaton::delta

/* Name:	hash
 * Class:	automaton
 * Purpose:	Computes a hash function on a state.
 * Parameters:	state_no	- (i) state number.
 * Returns:	Hash function for the state.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	The hash function is computed on outgoing transitions.
 *		It does not depend on the order of outgoing transitions.
 */
unsigned long int
automaton::hash(const stateno_t state_no) {
  const unsigned long int NTRANS_WEIGHT = 5623;
  const unsigned long int TARGET_WEIGHT = 6079;
  const unsigned long int LABEL_WEIGHT = 7243;
  const unsigned long int FINAL_WEIGHT = 3221;
  unsigned long int h = states[state_no].get_ntrans() * NTRANS_WEIGHT;
  stateno_t t = states[state_no].get_transitions();
  for (stateno_t t = states[state_no].get_transitions(); t != -1;
       t = transitions[t].get_next()) {
    h += (transitions[t].get_target() + 1) * TARGET_WEIGHT;
    h += transitions[t].get_label() * LABEL_WEIGHT;
    h += transitions[t].get_final() * FINAL_WEIGHT;
  }
  return h;
}//automaton::hash

/* Name:	register_find_or_put
 * Class:	automaton
 * Purpose:	Finds an equivalent state in the register
 *		or puts the state there.
 * Parameters:	state_no	- (i) state to be found or registered.
 * Returns:	Equivalent state or the same state if equivalent state
 *		not found.
 * Globals:	None.
 * Exceptions:	Standard std::vector and std::set exceptions.
 * Remarks:	The register is organized as a vector of sets of state numbers.
 */
stateno_t
automaton::register_find_or_put(const stateno_t state_no) {
#ifdef DEBUG
  std::cerr << "register_find_or_put(" << state_no << ")" << std::endl;
#endif  
  unsigned long int h = hash(state_no) % REGISTER_SIZE;;
#ifdef DEBUG
  std::cerr << "Hash is " << h << std::endl;
#endif  
  std::set<stateno_t>::iterator r;
#ifdef DEBUG
  std::cerr << "set content is [";
  std::copy(Register[h].begin(), Register[h].end(),
	    std::ostream_iterator<stateno_t>(std::cerr, " "));
  std::cerr << "]" << std::endl;
#endif  
  if ((r = Register[h].find(state_no)) != Register[h].end()) {
    // Equivalent state found
#ifdef DEBUG
    std::cerr << "Found " << *r << std::endl;
#endif    
    return *r;
  }
  else {
    // The state state_no is unique
#ifdef DEBUG
    std::cerr << "State is unique" << std::endl;
#endif    
    Register[h].insert(state_no);
    return state_no;
  }
}//automaton::register_find_or_put

/* Name:	repl_or_reg
 * Class:	None.
 * Purpose:	Follows a path of states determined by a state and a string
 *		replacing states with equivalent ones in the register
 *		or putting them into the register.
 * Parameters:	a		- (i/o) automaton to be processed;
 *		s		- (i/o) state at the start of a path;
 *		str		- (i) string to be followed.
 * Returns:	The state starting the path after minimization.
 * Globals:	None.
 * Exceptions:
 * Remarks:	The path is processed from the end of the path
 *		towards the start.
 *		Then the states are compared against states in the register.
 *		If a state is found equivalent to another state already present
 *		in the register, the state is replaced with the state
 *		in the register, and the transition is redirected towards it.
 *		If there is no equivalent state, the state is put
 *		into the register.
 */
stateno_t
repl_or_reg(automaton &a, const stateno_t s, const unsigned char *str) {
#ifdef DEBUG
  std::cerr << "repl_or_reg(a, " << s << ", " << str << ")" << std::endl;
#endif  
  if (*str) {
    // We are not at the end of the path
    a.change_transition(s, repl_or_reg(a, a.delta(s, *str), str + 1), *str);
  }
  return a.register_find_or_put(s);
}//repl_or_reg

/* Name:	remove_from_register
 * Class:	automaton
 * Purpose:	Removes a state from the register.
 * Parameters:	s		- (i) state number to be removed.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:	Standard std::vector and std::set exceptions.
 * Remarks:	None.
 */
void
automaton::remove_from_register(const stateno_t s) {
  Register[hash(s)].erase(s);
}//automaton::remove_from_register

/* Name:	find_equiv
 * Class:	automaton
 * Purpose:	Finds an equivalent state in the Register.
 * Parameters:	state_no	- (i) state number of the state, for which
 *					an equivalent state is to be found;
 *		fanout		- (i) number of initial outgoing transitions
 *					to skip.
 * Returns:	Equivalent state number or -1.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	The function is used for finding a small brother of a state.
 *		A small brother is a state with a set of outgoing transitions
 *		that is a subset of all transitions of the state state_no
 *		The transitions of the small brother need to be at the end
 *		of a list of transitions of the state state_no.
 *		To find a smaller brother, a state equivalent to it
 *		must first be created. Because it has a subset of transitions
 *		of the existing big brother, we do not need to create
 *		its transitions.
 */
stateno_t
automaton::find_equiv(const stateno_t state_no, const unsigned int skip) {
  // Get the number of transitions in state_no
  int bf = states[state_no].get_ntrans();
  stateno_t t = states[state_no].get_transitions();
  unsigned int skp = skip;
  do {
    t = transitions[t].get_next(); --skp;
  } while (skp > 0);
  // Create a state equivalent to the small brother
  stateno_t s = new_state();
  states[s].set_transitions(t);
  states[s].set_ntrans(states[state_no].get_ntrans() - skip);
  // Find an equivalent state
  stateno_t ss = -1;
  unsigned long int h = hash(s) % REGISTER_SIZE;
  std::set<stateno_t>::iterator r;
  if ((r = Register[h].find(s)) != Register[h].end()) {
    // Equivalent state found
    ss = *r;
  }
  // Remove the phony state
  states[s].set_transitions(-1);
  states[s].set_ntrans(0);
  delete_state(s);
  return ss;
}//automaton::find_equiv

/* Name:	add_str_sorted
 * Class:	None.
 * Purpose:	Adds string str to automaton a using the sorted construction.
 * Parameters:	a		- (i/o) automaton to be used;
 *		str		- (i) the string to be added;
 *		prev		- (i) previous string added to the automaton.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	None.
 */
void
add_str_sorted(automaton &a, const unsigned char *str,
	       const unsigned char *prev) {
#ifdef DEBUG
  std::cerr << str << " --- " << prev << std::endl;
#endif  
  const unsigned char *p = str;
  const unsigned char *pp = prev;
  stateno_t s = a.get_root();			// initial state
  // Traverse the longest common prefix
  stateno_t t = a.delta(s, *p);
  while (*p && t != -1) {
    s = t; p++; pp++; t = a.delta(s, *p);
  }//while
  // Minimize previously added suffix
  if (a.get_state(s).get_transitions() != -1) {
    a.change_transition(s, repl_or_reg(a, a.delta(s,*pp), pp + 1), *pp);
  }
  // Add the current suffix
  while (*p) {
    a.add_transition(s, t = a.new_state(), *p, p[1] == '\0'); s = t; p++;
  }
}//add_str_sorted

/* Name:	sorted_constr
 * Class:	None.
 * Purpose:	Constructs an automaton from a set of sorted strings.
 * Parameters:	a		- (i/o) automaton to construct;
 *		input		- (i) input stream containing the strings.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	The automaton should initially be empty.
 */
void
sorted_constr(automaton &a, std::istream &input) {
  std::string	input_line;
  std::string	prev_input_line;

  while(getline(input, input_line)) {
#ifdef DEBUG
    std::cerr << "'" << input_line << "'" << std::endl;
#endif
    if (input_line == "") continue;
    add_str_sorted(a, (unsigned char *)input_line.c_str(),
		   (unsigned char *)prev_input_line.c_str());
    prev_input_line = input_line;
  }
  input_line = "";
  add_str_sorted(a, (unsigned char *)input_line.c_str(),
		 (unsigned char *)prev_input_line.c_str());
}//sorted_constr

/* Name:	add_str_unsorted
 * Class:	None.
 * Parameters:	a		- (i/o) automaton to be used;
 *		str		- (i) the string to be added.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	None.
 */
void
add_str_unsorted(automaton &a, const unsigned char *str) {
  const unsigned char *p = str;
  std::vector<stateno_t> path;
  stateno_t s = a.get_root();			// initial state
  // Traverse the longest common prefix up to confluence states
  stateno_t t = a.delta(s, *p);
  while (*p && t != -1 && a.get_state(s).get_hit_count() == 1) {
    s = t; p++; t = a.delta(s, *p);
    path.push_back(s);
  }
  stateno_t lir = s;			// first state not in the register
  a.remove_from_register(s);
  // If the word is a prefix of a word that is already in the language
  // of the automaton, the next three loops will not run
  int change_made = 0;
  if (*p == '\0') {
    change_made = 1;
  }
  // Traverse confluence states in the longest common prefix
  while (*p && t != -1) {
    // Clone confluence states, clear finality if any
    // except when the word is a prefix of an existing word
    a.change_transition(s, t = a.new_state(t), *p, *p); s = t;
    path.push_back(s);
  }
  // Add the current suffix
  while (*p) {
    a.add_transition(s, t = a.new_state(), *p, p[1] == '\0'); s = t;
    path.push_back(s);
  }
  // Minimize back the suffix
  s = path.back(); path.pop_back(); --p;
  while (s != lir) {
    change_made = 0;
    stateno_t r = repl_or_reg(a, s, p);
    if (r != s) {
      // a.remove_from_register(path.top());
      change_made = 1;
      a.change_transition(path.back(), r, *p);
    }
    path.pop_back();
  };
  // As changes percolate towards the initial state,
  // continue updating states and the register,
  // as well as performing minimization
  while (change_made && path.size() > 0) {
    change_made = 0;
    stateno_t r = repl_or_reg(a, s, p);
    if (r != s) {
      a.remove_from_register(path.back());
      change_made = 1;
      a.change_transition(path.back(), r, *p);
    }
    path.pop_back();
  }
}//add_str_unsorted

/* Name:	unsorted_constr
 * Class:	None.
 * Purpose:	Constructs an sutomaton from a set of unsorted strings.
 * Parameters:	a		- (i/o) automaton to construct;
 *		input		- (i) input stream containing the strings.
 * Returns:	Nothing.
 * Exceptions:
 * Remarks:	The automaton should initially be empty.
 *		Input may constain duplicates.
 */
void
unsorted_construction(automaton &a, std::istream &input) {
  state s;
  std::basic_string<char>	input_line;
  std::basic_string<char>	prev_input_line;
  a.get_states()->push_back(s);		// the initial state

  while(getline(input, input_line)) {
    add_str_unsorted(a, (unsigned char *)input_line.c_str());
  }
}//unsorted_construction

/* Name:	number_entries
 * Class:	automaton
 * Purpose:	Assign cardinality of right language of a state to that state.
 * Parameters:	state_no	- (i) state number.
 * Returns:	Cardinality of state number state_no.
 * Globals:	None.
 * Exceptions:	Standard std::vector exceptions.
 * Remarks:	Visited states are marked with mark1 to avoid recounting.
 *		The cardinality is stored in entries vector
 *		index with state numbers.
 *		This is used for perfect hashing.
 *		This would not work for cyclic automata as they have
 *		infinite number of entries.
 */
long int
automaton::number_entries(const stateno_t state_no) {
  if (!states[state_no].is_mark1()) {
    states[state_no].set_mark1();
    //entries[state_no] = states[state_no].get_final();
    for (stateno_t t = states[state_no].get_transitions(); t != -1;
	 t = transitions[t].get_next()) {
      entries[state_no] += number_entries(transitions[t].get_target());
      entries[state_no] += transitions[t].get_final();
    }
  }
  return entries[state_no];
}//automaton::number_entries

/* Name:	bb_hash
 * Class:	None.
 * Purpose:	Calculates a hash function for a big brother sparse table.
 * Parameters:	bbv	- (i) sparse table;
 *		key	- (i) the key to the table.
 * Returns:	The hash value.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	None.
 */
stateno_t
bb_hash(const std::vector<bbrother_struct> &bbv, const stateno_t key) {
  return (key * 7243) % bbv.size();
}//bb_hash

/* Name:	bb_insert
 * Class:	None.
 * Purpose:	Inserts a big brother information into a sparse table.
 * Parameters:	bbv	- (i/o) the sparse table;
 *		bb	- the info.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	We assume there is always at least one empty place in the table.
 */
void
bb_insert(std::vector<bbrother_struct> &bbv, const bbrother_struct bb) {
  stateno_t h = bb_hash(bbv, bb.state_no);
  while (bbv[h].state_no != -1) {
    h = (h + 1) % bbv.size();
  }
  bbv[h] = bb;
}//bb_insert

/* Name:	bb_find
 * Class:	None.
 * Purpose:	Retrieves information about a big brother of a state.
 * Parameters:	bbv		- (i) big brother sparse table.
 *		state_no	- (i) state that has a big brother.
 * Returns:	A structure containing the big brother and an offset within it.
 * Globals:	None.
 * Exceptions:
 * Remarks:	We assume the state is stored in bbv, so it can always be found.
 */
bbrother_struct
bb_find(const std::vector<bbrother_struct> &bbv, const stateno_t state_no) {
  stateno_t h = bb_hash(bbv, state_no);
  while (bbv[h].state_no != state_no) {
    h = (h + 1) % bbv.size();
  }
  return bbv[h];
}//bb_find

/* Name:	share_arcs
 * Class:	automaton
 * Purpose:	Finds states that can be stored inside other states.
 * Parameters:	None.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	Smaller states (states with fewer outgoing transitions)
 *		can only be stored inside bigger states.
 *		Vector nnstates holds lists of states sorted on their number of
 *		outgoing transitions. Each item has a state (state_no),
 *		and an index of the next item on the list (-1 means null).
 *		Vector nstates holds heads of lists for states with the number
 *		of outgoing transitions as its index.
 *		Vector bbrothers holds addresses of bigger states
 *		that have a suite of outgoing transitions that include
 *		all outgoing transitions of a smaller state.
 *		It has three fields: an address of a small brother,
 *		an address of the big brother, and the
 *		number of initial transitions to skip in it.
 *		It is stored as a sparse vector.
 *		To avoid sorting states on their number of outgoing transitions,
 *		we group the states into lists. The nnstates vector holds
 *		the next state number on the list, i.e. nnstate[x]
 *		is the state number (and also and index in nnstates)
 *		of the next state having the same number
 *		of outgoing transitions.
 *		Vector nstates (single n) holds heads of such lists,
 *		i.e. nstates[x] holds the beginning of a list of states
 *		having x outgoing transitions.
 */
void
automaton::share_arcs(void) {
  if (options & SORT_ON_FREQ) {
    // Destroy the register, sort transitions, and repopulate the register
    sort_transitions();
  }
  // Group states according to the number of their outgoing transitions
  // and arrange them in lists in nnstates with heads in nstates
  int max_trans = -1;
  stateno_t states_counted = 0;
  std::vector<stateno_t> nstates;
  stateno_t nnstates[number_of_states];
  for (stateno_t i = 0; i < states.size(); i++) {
    if (!states[i].is_deleted()) {
      long int nn = states[i].get_ntrans();
      while (nn > max_trans) {
	nstates.push_back(-1); max_trans++;
      }
      nnstates[i] = nstates[nn];
      nstates[nn] = i;
      states_counted++;
    }
  }
  stateno_t bbv_size = std::max((stateno_t)1000, number_of_states / 20);
  bbrother_struct bbi = bbrother_struct(-1, -1, 0);
  bbrother = std::vector<bbrother_struct>(bbv_size, bbi);
  number_of_bbrothers = 0;
  for (int i = max_trans; i > 1; --i) {
    // For all possible decreasing numbers `i' of outgoing transitions
    for (stateno_t s = nstates[i]; s != -1; s = nnstates[s]) {
      // For all states having the same number of outgoing transitions
      for (int k = 1; k < i; k++) {
	// For decreasing numbers of transitions to skip to get smaller brother
	stateno_t sb = 0;
	if ((sb = find_equiv(s, k)) != -1) {
	  // Equivalent transition found
	  if (!states[sb].is_small_brother()) {
	    // If the state found has no big brother itself
	    // Store the address of the big brother in bbrother vector
	    //bbrother[number_of_bbrothers].bb = nnstates[s].state_no;
	    // Store how many transitions in big brother to skip
	    //bbrother[number_of_bbrothers].offset = k;
	    bb_insert(bbrother,bbrother_struct(sb, s, k));
	    // Mark the states
	    states[sb].set_small_brother_mark();
	    states[s].set_big_brother_mark();
	  }//if not marked as already stored somewhere else
	}//if smaller brother found
      }//for all possible offsets
    }//for all states with i outgoing transitions
  }//for all outgoing transition numbers
}//automaton::share_arcs

/* Name:	cmp_tr
 * Class:	None.
 * Purpose:	Compares two transitions.
 * Parameters:	t1	- (i) first transition;
 *		t2	- (i) second transition;
 *		f	- (i) frequency table.
 * Returns:	-1 if t1.label < t2.label,
 *		0 if t1.label = t2.label,
 *		1 if t1.label > t2.label.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	Parameter `f' is a frequency table for all labels indexed
 *		by their code (0-255).
 */
int
cmp_tr(const void *t1, const void *t2, void *f) {
  return ((unsigned long int *)f)[((const transition *)t1)->get_label()] -
    ((unsigned long int *)f)[((const transition *)t2)->get_label()];
}//cmp_tr


/* Name:	cmp_tr_rev
 * Class:	None.
 * Purpose:	Compares two transitions.
 * Parameters:	t1	- (i) first transition;
 *		t2	- (i) second transition;
 *		f	- (i) frequency table.
 * Returns:	1 if t1.label < t2.label,
 *		0 if t1.label = t2.label,
 *		-11 if t1.label > t2.label.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	Parameter `f' is a frequency table for all labels indexed
 *		by their code (0-255).
 */
int
cmp_tr_rev(const void *t1, const void *t2, void *f) {
  return ((unsigned long int *)f)[((const transition *)t2)->get_label()] -
    ((unsigned long int *)f)[((const transition *)t1)->get_label()];
}//cmp_tr_rev


/* Name:	sort_transitions
 * Class:	automaton
 * Purpose:	Sorts all outgoing transitions of states based on the frequency
 *		of their labels.
 * Parameters:	None.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	There is no need to do anything with the Register, as the hash
 *		function does not depend of the order of the transitions.
 */
void
automaton::sort_transitions(void) {
  // Calculate frequency of transitions
  unsigned long int freq[256];
  memset(freq, 0, 256);
  for (stateno_t i = 0; i < states.size(); i++) {
    // For all states
    if (!states[i].is_deleted()) {
      // that are active
      for (stateno_t j = states[i].get_transitions(); j != -1;
	   j = transitions[j].get_next()) {
	// For all transitions
	freq[transitions[j].get_label()]++;
      }
    }
  }
  // Sort transitions according to frequency of their labels
  for (stateno_t i = 0; i < states.size(); i++) {
    if (!states[i].is_deleted()) {
      // Gather transitions
      transition tr[256];	// there cannot be more transitions for a state
      int k = 0;		// stored transition number per state
      for (stateno_t j = states[i].get_transitions(); j != -1;
	   j = transitions[j].get_next()) {
	tr[k++] = transitions[j];
      }
      if (options & DESCENDING) {
	qsort_r((void *)tr, 256, sizeof(transition), cmp_tr_rev, freq);
      }
      else {
	qsort_r((void *)tr, 256, sizeof(transition), cmp_tr, freq);
      }
      // Store sorted transitions
      k = 0;
      for (stateno_t j = states[i].get_transitions(); j != -1;
	   j = transitions[j].get_next(), k++) {
	transitions[j].set_label(tr[k].get_label());
	transitions[j].set_target(tr[k].get_target());
	transitions[j].set_final(tr[k].get_final());
      }
    }
  }
}//sort_transitions

/* Name:	count_next
 * Class:	automaton
 * Purpose:	Count the number of transitions that have the NEXT flag set.
 * Parameters:	state_no	- (i) state number from which to begin.
 * Returns:	The number of transitions with the NEXT flag set in the part
 *		of the automaton starting from state state_no that have not
 *		yet been counted.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	We assume that assigning addresses and writing the transitions
 *		follows the same paths in the automaton.
 *		We try to maximize the number of NEXT flags.
 *		The NEXT flag is set when the target of the last transition
 *		of the state is stored directly after that transition.
 *		Therefore, we visit the target of the last transition first.
 *		The remaining transitions are visited from the first
 *		to the penultimate one.
 *		Maximization is done by reordering transitions.
 *		If the state has a big brother, i.e. it is stored inside
 *		another state (the big brother), the big brother is used
 *		instead of the current state.
 *		If the last transition has a target to a state that is already
 *		stored, we reorder transitions only if the current state
 *		is not marked as a big brother, as reordering transitions
 *		of a big brother could change the transition set of the small
 *		brother stored within it.
 *
 *		The last outgoing transition of a state can get the NEXT flag
 *		if:
 *		1. The target state has outgoing transitions (it is not
 *			written as 0).
 *		2. The target state has not been written (visited in
 *			count_next), i.e. it has no cnxt flag.
 *		3. The target state is not a small brother (as it is kept
 *			inside another state that starts earlier).
 *
 *		Reordering the transitions is possible when:
 *		1. There are at least 2 transitions, and the last one
 *			cannot have the NEXT flag.
 *		2. The state is not marked as big brother or small brother,
 *			as its order of transitions would be fixed.
 *		3. One transition that is not last meats requirements given
 *			above.
 */
stateno_t
automaton::count_next(const stateno_t state_no) {
  stateno_t n = 0;		// counter for NEXT flags
  if (states[state_no].is_cnxt() || states[state_no].get_ntrans() == 0) {
    // The state has already been visited (and its transitions counted)
    // or the state has no transitions (state number 0)
    return 0;
  }
  else {
    states[state_no].set_cnxt(); // mark the state as visited
    if (states[state_no].is_small_brother()) {
      // The state state_no is stored entirely inside another state
      // Go to that other state to follow addressing scheme
      bbrother_struct bbs = bb_find(bbrother, state_no);
      n += count_next(bbs.bb);
    }
    else if (states[state_no].get_ntrans() > 1 &&
	     !states[state_no].is_big_brother()) {
      // Reordering possible if needed
      // Get the last transition
      stateno_t tt = states[state_no].get_transitions();
      stateno_t prev1 = -1;
      stateno_t prev2 = -1;
      while (transitions[tt].get_next() != -1) {
	prev2 = tt;
	tt = transitions[tt].get_next();
      }
      // Now tt is the last transition
      // See if the target of the last transition is not visited
      stateno_t t1 = transitions[tt].get_target();
      if (!states[t1].can_be_next()) {
	// The target of the last transition is already stored
	// or the pointer of the transition is zero
	for (stateno_t t = states[state_no].get_transitions(); t != tt;
	     t = transitions[t].get_next()) {
	  // Try reordering transitions to set the NEXT flag
	  stateno_t ttt = transitions[t].get_target();
	  if (states[ttt].can_be_next()) {      // target not yet stored
	    // Swap transition t and tt (the last one)
	    // Next field of tt should point to next of t unless tt comes
	    // right after t, then it points to t
	    transitions[tt].set_next(transitions[t].get_next() == tt ? t
				     : transitions[t].get_next());
	    //
	    if (transitions[t].get_next() != tt) {
	      transitions[prev2].set_next(t);
	    }
	    // Now t becomes the last transition
	    transitions[t].set_next(-1);
	    if (prev1 != -1) {
	      // tt now takes the place of t
	      transitions[prev1].set_next(tt);
	    }
	    else {
	      // As t was the first transition, tt becomes the first transition
	      states[state_no].set_transitions(tt);
	    }
	    tt = t;
	    states[state_no].set_next_flag();
	    break;
	  }
	  prev1 = t;
	}
      }
      // Possibly after reordering of transitions
      stateno_t ttt = transitions[tt].get_target();
      if (states[ttt].can_be_next()) {
	// The last transition will have the NEXT flag, so
	// the target of the last transition will be stored first
	transitions[tt].set_next_flag(1);
	states[state_no].set_next_flag();
	stateno_t n = 1 + count_next(ttt);
	for (stateno_t t = states[state_no].get_transitions(); t != tt;
	     t = transitions[t].get_next()) {
	  ttt = transitions[t].get_target();
	  n += count_next(ttt);
	}
	return n;
      }
      else {
	// No transition has the NEXT flag
	stateno_t n = 0;
	for (stateno_t t = states[state_no].get_transitions(); t != -1;
	     t = transitions[t].get_next()) {
	  stateno_t ttt = transitions[t].get_target();
	  n += count_next(ttt);
	}
	return n;
      }
    }
    else {
      // No reordering of transitions is possible, but NEXT flag still possible
      stateno_t tt = -1;
      stateno_t n = 0;
      for (stateno_t t = states[state_no].get_transitions(); t != -1;
	   t = transitions[t].get_next()) {
	tt = t;
      }
      // Now tt is the last transition
      stateno_t ttt = transitions[tt].get_target();
      if (states[ttt].get_ntrans() > 0 && states[ttt].can_be_next()) {
	states[state_no].set_next_flag();
	transitions[tt].set_next_flag(1);
	n += 1 + count_next(ttt);
	for (stateno_t t = states[state_no].get_transitions(); t != tt;
	     t = transitions[t].get_next()) {
	  n += count_next(transitions[t].get_target());
	}
	return n;
      }
      else {
	for (stateno_t t = states[state_no].get_transitions(); t != -1;
	     t = transitions[t].get_next()) {
	  n += count_next(transitions[t].get_target());
	}
	return n;
      }
    }
  }
}//automaton::count_next

/* Name:	bit_length
 * Class:	None.
 * Purpose:	Counts the length of a number in bits.
 * Parameters:	n		- (i) the number.
 * Returns:	The number of bits needed to represent the number.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	None.
 */
unsigned int
bit_length(const unsigned long int n) {
  unsigned long int nn = n;
  unsigned int counter = 0;
  while (nn) {
    nn >>= 1; counter++;
  }
  return counter;
}//bit_length

/* Name:	count_arcs
 * Class:	automaton
 * Purpose:	Calculates how much memory the automaton will occupy.
 * Parameters:	None.
 * Returns:	The number of bytes occupied by the compressed automaton.
 * Globals:	None.
 * Exceptions:
 * Remarks:	This is needed to learn the length of pointers
 *		in the compressed automaton.
 *		The size depends on the number of transitions,
 *		on the number of states contained in other states,
 *		on the number of transitions with the NEXT flag
 */
stateno_t
automaton::count_arcs(void) {
  entryl = 0;
  if (options & WITH_NUMBERS) {
    entries = new stateno_t[transitions.size()];
    entryl = bits_in(number_entries(root))/8 +1;
  }
  number_of_bbrothers = 0;
  if (options & OPTIMIZE) {
    share_arcs();
  }
  stateno_t n = count_next(get_superroot());	// count arcs with NEXT flag set
  gtl = 0;
  unsigned int new_gtl = 0;
  stateaddr_t aut_size = 0;
  do {
    gtl++;
    aut_size = (number_of_transitions - number_of_bbrothers)
      * (gtl + 1) - n * (gtl - 1) + number_of_states * entryl;
    new_gtl = ((bits_in(aut_size) + flags_size) / 8) + 1;
  } while (gtl != new_gtl);
  return aut_size;
}//automaton::count_arcs

/* Name:	compute_addresses
 * Class:	automaton
 * Purpose:	Computes addresses of states in the automaton.
 * Parameters:	state_no	- (i) state number to begin with;
 * Returns:	The address of the state state_no.
 * Globals:	current_loc	- (i) first available location address.
 * Exceptions:
 * Remarks:	Note that the null state at the beginning has a transition
 *		that leads nowhere (or rather to itself), so it should
 *		be handled separately.
 *		Visiting and storing states should be done in the same order
 *		as in count_next().
 *		We need to know the addresses of states before we write
 *		the automaton to a file, because they are used as target
 *		addresses of transitions. Transitions lead to states
 *		that possibly are yet to be written.
 *		Addresses are stored in stateaddr vector that must already
 *		be allocated.
 */
stateaddr_t
automaton::compute_addresses(const stateno_t state_no) {
  stateaddr_t addr = 0;
  if (states[state_no].get_ntrans() == 0) {
    stateaddr[state_no] = 0;
    return 0;
  }
  if (states[state_no].is_small_brother()) {
    // The state will be stored inside another state
    // Its address is the address of the big brother plus the size
    // of initial additional transitions of the big brother
    bbrother_struct bbs = bb_find(bbrother, state_no);
    addr = compute_addresses(bbs.bb) + (gtl + 1) * bbs.offset;
    stateaddr[state_no] = addr;
    return addr;
  }
  else {
    if (!states[state_no].is_addrcmpt()) {
      // The address of this state is not yet computed
      states[state_no].set_addrcmpt();
      addr = current_loc;
      stateaddr[state_no] = addr;
      current_loc += states[state_no].get_ntrans() * (gtl + 1);
      stateno_t t = states[state_no].get_transitions();
      stateno_t tt = t;
      if (states[state_no].has_next_flag()) {
	// The last transition of the state holds the NEXT flag
	current_loc -= (gtl - 1); // the last transition is shorter
	// get the last transition
	while (t != -1) {
	  // Go to the last transition (if any)
	  tt = t;		// store the index of the last transition in tt
	  t = transitions[t].get_next();
	}
	// tt is now the last transition
	// First go to the target of the last transition
	compute_addresses(transitions[tt].get_target());
	// Then compute the addresses of the targets of the remaining
	// transitions
	for (t = states[state_no].get_transitions(); t != tt;
	     t = transitions[t].get_next()) {
	  compute_addresses(transitions[t].get_target());
	}
      }
      else {
	while (t != -1) {
	  compute_addresses(transitions[t].get_target());
	  t = transitions[t].get_next();
	}

      }
      return addr;
    }
    else {
      return stateaddr[state_no];
    }
  }
}//automaton::compute_addresses

/* Name:	serialize_int
 * Class:	None.
 * Purpose:	Writes a long integer as a sequence of bytes.
 * Parameters:	buffer		- (o) where to write the bytes;
 *		n		- (i) number to serialize;
 *		len		- (i) number of bytes to represent the integer.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:	FSACL_NumberBigger if the number need more than len bytes.
 * Remarks:	There must be at least len bytes available in the buffer.
 *		Little endian is used, i.e. the least significant part
 *		of the number comes first.
 */
void
serialize_int(char *buffer, const unsigned long int n, unsigned int len) {
  unsigned long int num = n;
  for (int i = 0; i < len; i++) {
    buffer[i] = num & 0xff;
    num = (num >> 8);
  }
  if (num) {
    throw(ErrorException(FSACL_NumberBigger));
  }
}//serialize_int

/* Name:	write_automaton
 * Class:	automaton
 * Purpose:	Writes a compressed automaton to a stream.
 * Parameters:	output		- (o) output stream.
 * Results:	None.
 * Globals:	None.
 * Exceptions:
 * Remarks:	The format of the automaton depends on the parameters.
 *		In all formats used in the library, states are implicit.
 *		They are represented as lists or sets of transitions.
 *		There are two major formats:
 *		1. Superimposed coding based on sparse matrix representation.
 *		   Transitions are of fixed length. A state is a set
 *		   of transitions. For each state, there is a fixed slot
 *		   for each transition with a given label. The slot is indexed
 *		   by the label. The slot can be filled or it can be empty.
 *		   A filled slot has a correct label corresponding to its index.
 *		   As usually most slots are empty, they can be used by other
 *		   states. This is implemented so that the distance between
 *		   two consecutive states is smaller than the size
 *		   of the alphabet. When a transition at a certain index
 *		   has a label different from its index, then it belongs
 *		   to another state.
 *		2. Lists of transitions. Transitions are stored one after
 *		   another. The address of a state is the address of its
 *		   first transition. The last transition of a state has
 *		   a marker (a flag). Transitions can be variable length.
 *		   There are a few variants of this format.
 */
void
automaton::write_automaton(std::ostream &output) {
  write_header(output);
  write_transition_vector(output);
}//automaton

/* Name:	write_header
 * Class:	automaton
 * Purpose:	Writes a header of an automaton file.
 * Parameters:	output		- (o) where to write the header to.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	The format of the header is:
 *		0:	\\	\
 *		1:	f	 +--- magic number
 *		2:	s	 |
 *		3:	a	/
 *		4:	fsa format version number (see INSTALL file)
 *		5:	character used as filler (see README file)
 *		6:	annotation separator - separates word form
 *			from annotations and possibly some other info
 *		7:	length of the pointer field in transitions (arcs)
 *			measured in bytes; when the automaton
 *			has numbered entries, then the pointer length
 *			occupies only the 4 lower bits, while
 *			the upper 4 bits are taken by the length in bytes
 *                      of the field that holds the cardinality
 *                      of the right language of the state, i.e. the number
 *                      of strings recognized from the state on chains of
 *                      transitions leading to all final states reachable
 *                      from the state
 *		8:	vector of transitions
 */
void
automaton::write_header(std::ostream &output) {
  signature sign_arc;
  sign_arc.sig[0] = '\\';
  sign_arc.sig[1] = 'f';
  sign_arc.sig[2] = 's';
  sign_arc.sig[3] = 'a';
  sign_arc.ver = 5;
  sign_arc.filler = filler;
  sign_arc.annot_sep = annot_separator;
  sign_arc.gtl = gtl;
  if (options & WITH_NUMBERS) {
    sign_arc.gtl += (entryl << 4);
  }
  output.write((char *)&sign_arc, sizeof(sign_arc));
}//automaton::write_header

/* Name:	write_transition_vector
 * Class:	automaton
 * Purpose:	Writes to a file the automaton as a sequence of transitions.
 * Parameters:	output		- (o) where to write the transitions to.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	Transitions are written in the same order as they had been
 *		numbered.
 */
void
automaton::write_transition_vector(std::ostream &output) {
  // Write the sink state
  char buffer[64];
  // The sink state has one transition with a null label, a gtl-long
  // pointer, a non-final flag, a cleared NEXT flag, and a raised STOP flag
  if (options & WITH_NUMBERS) {
    serialize_int(buffer, 0, entryl); // Write the number of entries
    output.write(buffer,entryl);
  }
  buffer[0] = '\0';		// label
  for (int i = 0; i < gtl; i++) {
    buffer[i + 1] = '\0';	// target pointer
  }
  output.write(buffer, gtl + 1);
  write_state(output, get_superroot());
}//automaton::write_transition_vector

/* Name:	serialize_transition
 * Class:	automaton
 * Purpose:	Writes a representation of a transition to a character buffer.
 * Parameters:	buffer		- (o) buffer to write to;
 *		t		- (i) transition to write;
 *		start		- (i) starting character in the buffer
 * Returns:	The next available byte offset in the buffer.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	It is assumed that there is sufficient room in the buffer
 *		for a representation of the transition.
 */
stateaddr_t
automaton::serialize_transition(char *buffer, const transition &t,
		     const unsigned int start) {
  unsigned int pp = start;
  buffer[pp++] = t.get_label();
  // Store the target address
  stateaddr_t ptr = stateaddr[t.get_target()];
  buffer[pp] = ((ptr & 0x1f) << 3);
  ptr >>= 5;
  // set flags
  buffer[pp] |= t.get_final(); // set FINAL flag
  if (t.get_next() == -1) {
    // Set the STOP flag
    buffer[pp] |= 2;
  }
  if (t.get_next_flag()) {
    buffer[pp++] |= 4;
  }
  else {
    pp++;
    for (int i = 1; i < gtl; i++) {
      buffer[pp++] = (ptr & 0xff);
      ptr >>= 8;
    }
  }
  return pp;
}//automaton::serialize_transition

/* Name:	write_state
 * Class:	automaton
 * Purpose:	Writes a state as a sequence of transitions to the output.
 * Parameters:	output		- (o) where to write the state to;
 *		state_no	- (i) state number.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:
 * Remarks:	Also states accessible from the state are written if possible.
 */
void
automaton::write_state(std::ostream &output, const stateno_t state_no) {
  char buffer[1024];
  unsigned int pp = 0;
  if (states[state_no].get_ntrans() == 0) {
    // This state is represented as state number 0
    return;
  }
  if (!states[state_no].is_written()) {
    states[state_no].set_written();
    if (states[state_no].is_small_brother()) {
      bbrother_struct bbs = bb_find(bbrother, state_no);
      write_state(output, bbs.bb);
      return;
    }
    if (options & WITH_NUMBERS) {
      // Write the number of entries
      serialize_int(buffer, entries[state_no], entryl);
      pp += entryl;
    }
    // Write the transitions of the state
    stateno_t tt = states[state_no].get_transitions();
    for (stateno_t t = tt; t != -1; t = transitions[t].get_next()) {
      // Write the transition
      pp = serialize_transition(buffer, transitions[t], pp);
    }
    output.write(buffer, pp);
    if (states[state_no].has_next_flag()) {
      stateno_t ttt = tt;
      for (stateno_t t = tt; t != -1; t = transitions[t].get_next()) {
	ttt = t;
      }
      write_state(output, transitions[ttt].get_target());
      for (stateno_t t = tt; t != ttt; t = transitions[t].get_next()) {
	write_state(output, transitions[t].get_target());
      }
    }
    else {
      for (stateno_t t = tt; t != -1; t = transitions[t].get_next()) {
	write_state(output, transitions[t].get_target());
      }
    }
  }
}//automaton::write_state

/* Name:	reduce_inner
 * Class:	automaton
 * Purpose:	Removes states and transitions in the automaton that all lead
 *		to the same set of annotations.
 * Parameters:	state_no	- (i) state to start from.
 * Returns:	The same state, or a state being the target of all transitions
 *		leaving state state_no.
 * Globals:	None.
 * Exceptions:	Standard memory exceptions.
 * Remarks:	This is used fro producing guessing automata.
 *		See my paper Treatment of Unknown Words, Workshop
 *		on Implementation of Automata WIA'98, Potsdam, Germany.
 *		The following rules apply:
 *		R1. The pruning process does not apply to states and transitions
 *			belonging to annotations.
 *		R2. A transition can be removed only if the pruning process
 *			has already visited all transitions that can be reached
 *			through the target state, except the transitions
 *			representing annotations, i.e. states are visited
 *			in postorder.
 *		R3. A transition cannot be removed if the target state has
 *			a transition that does not belong to annotations,
 *			but cannot be removed, i.e. transitions of the target
 *			state distinguish between different annotations.
 *		R4. A state can always be replaced with an equivalent one.
 *		R5. A state with all transitions leading to one state
 *			can be removed with all transitions, and transitions
 *			that point to it should be redirected
 *			to the target state.
 *
 *		As a result of processing, the state can:
 *		1. Stay the same.
 *		2. Be modified, but without the need to be deleted or replaced.
 *		3. Be modified, so that it becomes equivalent to another
 *			state and should be replaced with an equivalent one.
 *		4. Be replaced with the target, when all its transitions
 *			lead to the same target.
 *
 *		In cases 3 and 4, the state cannot be deleted if it has
 *		more than one incoming transition. However, in each case,
 *		we can process the state once, and mark the state so that
 *		we know what to do when we visit the state again. When the state
 *		is to be deleted, we decrease the number of its incoming
 *		transitions until it reaches one, and then delete it.
 *
 *		The marking is as follows:
 *		Case 1 and 2 - reduced.
 *		Case 3 - redundant.
 *		case 4 - waste.
 *
 *		
 */
stateno_t
automaton::reduce_inner(const stateno_t state_no) {
  if (states[state_no].get_ntrans() == 0) {
    // No transitions or invalid annotation separator
    throw(ErrorException(FSACL_InvSepAn));
  }
  bool has_annotations = false;
  bool reducible = true;
  if (!states[state_no].is_processed()) {
    // The state is visited in reduce_inner() for the first time
    bool state_modified = false;
    for (stateno_t tt = states[state_no].get_transitions(); tt != -1;
	 tt = transitions[tt].get_next()) {
      if (transitions[tt].get_label() != annot_separator) {
	// Reduce children
	stateno_t new_target = reduce_inner(transitions[tt].get_target());
	if (new_target != transitions[tt].get_target()) {
	  // We have changed at least this target
	  if (!state_modified) {
	    // The first change, so we must remove state_no from the Register
	    Register[hash(state_no)].erase(state_no);
	    state_modified = true; // Do not remove it from the register again!
	  }
	  transitions[tt].set_target(new_target);
	}
      }
      else {
	has_annotations = true;
      }
    }
    if (!has_annotations) {
      // The state has no annotation separator labeled transitions,
      // so it can be reduced
      // Check whether all transitions have the same target
      stateno_t d = transitions[states[state_no].get_transitions()].get_target();
      for (stateno_t tt = states[state_no].get_transitions(); tt != -1;
	 tt = transitions[tt].get_next()) {
	if (transitions[tt].get_target() != d) {
	  reducible = false;	// transitions differ
	}
      }
    }
    // Set markers
    if (reducible) {
      states[state_no].set_waste();
    }
    else if (state_modified) {
      if (register_find_or_put(state_no) != -1) {
	states[state_no].set_redundant();
      }
      else {
	states[state_no].set_reduced();
      }
    }
  }
  // The state has been processed (either previously or now)
  if (states[state_no].is_reduced()) {
    // The state is not changed
    return state_no;
  }
  else if (states[state_no].is_redundant()) {
    // The state should be replaced with an equivalent one
    stateno_t r = -1;
    if ((r = register_find_or_put(state_no)) != -1) {
      // The equivalent state found
      states[r].hit();		// We delete state state_no,
                                // avoid deleting children
      delete_useless_state(state_no);
    }
    else {
      states[state_no].clear_redundant();
      states[state_no].set_reduced();
      return state_no;
    }
  }
  else if (states[state_no].is_waste()) {
    stateno_t tg = transitions[states[state_no].get_transitions()].get_target();
    states[tg].hit();
    delete_useless_state(state_no);
    return tg;
  }
  else {
    std::cerr << "Error! No mark in reduce_inner!\n";
    exit(1);
  }
}//automaton::reduce_inner

/* Name:	prune_arcs
 * Class:	automaton
 * Purpose:	Removes parts of the automaton to make guessing better
 *		and faster.
 * Parameters:	state_no	- (i) the state to start with.
 * Returns:	The state after modifications if applicable.
 * Globals:	None.
 * Exceptions:	Memory management exceptions.
 * Remarks:	The rules:
 *		1. If node n has an arc labeled with the annotation separator
 *		   it cannot be pruned.
 *		2. The number of arcs to be pruned must be at least
 *		   MIN_PRUNED for one node.
 *		3. There should be exactly one arc leading to a node
 *		   other than those pointed to by the nodes to be pruned.
 *		4. The arcs to be pruned must all lead to the same node
 *		   that must have all arcs leaving that node labeled with
 *		   annotation separator.
 *		4. All pruned arcs are replaced with children of the node
 *		   they lead to.
 *
stateno_t
automaton::prune_arcs(const stateno_t state_no) {
  const unsigned int MIN_PRUNE = 2;
  struct prune_struct {
    stateno_t		dest;
    unsigned int	count;
    prune_struct(const stateno_t d) : dest(d), count(1) {}
  };

  if (states[state_no].get_ntrans() <= MIN_PRUNE) {
    // Nothing to be pruned - too few transitions
    return state_no;
  }
  // Collect different destinations of transitions
  std::vector<prune_struct> children;
  for (stateno_t tt = states[state_no].get_transitions(); tt != -1;
       tt = transitions[tt].get_next()) {
    bool found = false;
    for (unsigned int i = 0; i < children.size(); i++) {
      if (children[i].dest == transitions[tt].get_target()) {
	children[i].count++;
	found = true;
      }
    }
    if (!found) {
      children.push_back(prune_struct(transitions[tt].get_target()));
    }
  }
  // Find the most frequent destination
  unsigned int biggest = 0;
  unsigned int biggest_index = 0;
  for (unsigned int i = 0; i < children.size(); i++) {
    if (children[i].count > biggest) {
      biggest = children[i].count;
      biggest_index = i;
    }
  }
  if (biggest < MIN_PRUNE * (states[state_no].get_ntrans() - biggest)) {
    // Not enough transitions to the most frequent destination, leave unchanged
    return state_no;
  }
  // Prune the state
  Register.erase(state_no);	// get it out the register
  stateno_t d = children[biggest_index].dest;
  stateno_t s = new_state();					   // new state
  stateno_t t = new_transition(states[state_no].get_transitions(), // in front
			       d, // put the most frequent destination
			       annot_separator, // with the annotation separator
			       0);		// not final
  states[s].set_transitions(t);			// new transition in front
  for (stateno_t tt = transitions[t].get_next(); tt != -1) {
    // Skip transitions to the most frequent destination
    while (tt != -1 && transitions[tt].get_target() == d) {
      tt = transitions[tt].get_next();
    }
  }
  // Delete the old state
  }//automaton::prune_arcs*/

/* Name:	write_any_construction_file
 * Class:	automaton
 * Purpose:	Optimizes automaton and writes it to a file.
 * Parameters:	output		- (o) output stream.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:	Throws ErrorException class with an integer as parameter.
 *		See fsacl.hh for a list.
 * Remarks:	Stream output must be open for writing.
 */
void
automaton::write_any_construction_file(std::ostream & output) {
  if (options & MAKE_INDEX) {
    root = reduce_inner(root);
  }
  stateno_t size = count_arcs();
  current_loc = gtl + 1;	// skip the null state
  stateaddr = new stateaddr_t[states.size()];
  compute_addresses(get_superroot());
  write_automaton(output);
}//automaton::write_any_construction_file

/* Name:	write_sorted_construction_file
 * Class:	None.
 * Purpose:	Constructs an automaton from sorted data, and writes it
 *		to a file.
 * Parameters:	input		- (i) input file name
 *					or empty for standard input;
 *		output		- (i) output file name
 *					or empty for standard output;
 *		options		- (i) options for building the automaton;
 *		annot_sep	- (i) annotation separator;
 *		filler		- (i) filler character;
 *		version		- (i) representation format version;
 * Returns:	Nothing.
 * Globals:	fsacl_errno	- (o) error code (0 is OK).
 * Exceptions:	None.
 * Remarks:	Invoke fsacl_get_errno right afterwards.
 */
void
write_sorted_construction_file(const char *input,
			       const char *output,
			       const unsigned int options,
			       const char annot_sep,
			       const char filler,
			       const unsigned int version) {
  fsacl_set_errno(0);
  try {
    automaton a = automaton();
    less_states l;
    initialize_Register(a,l);

    std::istream *ifn;
    std::ostream *outs;
    std::ifstream inpf;
    std::ofstream outf;
    if (input) {
      inpf.open(input, std::ifstream::in);
      if (!inpf) {
	throw(ErrorException(FSACL_InvInpFile));
      }
      ifn = &inpf;
    }
    else {
      ifn = &std::cin;
    }
    if (output) {
      outf.open(output, std::ios::binary); // flag for M$ bug
      if (!outf) {
	throw(ErrorException(FSACL_InvOutFile));
      }
      outs = &outf;
    }
    else {
      outs = &std::cout;
    }

    a.set_option(options);
    a.set_annot_separator(annot_sep);
    a.set_filler(filler);
    sorted_constr(a, *ifn);
    a.write_any_construction_file(*outs);
  }
  catch (ErrorException e) {
    fsacl_set_errno(e.error_code);
  }
  catch (...) {
    fsacl_set_errno(FSACL_OtherErr);
  }
}//write_sorted_construction_file

/* Name:	write_unsorted_construction_file
 * Class:	None.
 * Purpose:	Constructs an automaton from unsorted data, and writes it
 *		to a file.
 * Parameters:	input		- (i) input file name
 *					or empty for standard input;
 *		output		- (o) output file name
 *					or empty for standard output;
 *		options		- (i) options for building the automaton;
 *		annot_sep	- (i) annotation separator;
 *		filler		- (i) filler character;
 *		version		- (i) representation format version.
 * Returns:	Nothing.
 * Globals:	fsacl_errno	- (o) error code (0 is OK).
 * Exceptions:	None.
 * Remarks:	Use fsacl_get_errno right afterwards.
 */
void
write_unsorted_construction_file(const char *input,
				 const char *output,
				 const unsigned int options,
				 const char annot_sep,
				 const char filler,
				 const unsigned int version) {
  fsacl_set_errno(0);
  try {
    automaton a = automaton();
    less_states l;
    initialize_Register(a,l);

    std::istream *ifn;
    std::ostream *outs;
    std::ifstream inpf;
    std::ofstream outf;
    if (input) {
      inpf.open(input, std::ifstream::in);
      if (!inpf) {
	throw(ErrorException(FSACL_InvInpFile));
      }
      ifn = &inpf;
    }
    else {
      ifn = &std::cin;
    }
    if (output) {
      outf.open(output, std::ios::binary); // flag for M$ bug
      if (!outf) {
	throw(ErrorException(FSACL_InvOutFile));
      }
      outs = &outf;
    }
    else {
      outs = &std::cout;
    }

    a.set_option(options);
    a.set_annot_separator(annot_sep);
    a.set_filler(filler);
    unsorted_construction(a, *ifn);
    a.write_any_construction_file(*outs);
  }
  catch (ErrorException e) {
    fsacl_set_errno(e.error_code);
  }
  catch (...) {
    fsacl_set_errno(FSACL_OtherErr);
  }
}//write_unsorted_construction_file


/* Name:	fsacl_set_errno
 * Class:	None.
 * Purpose:	Sets an error code.
 * Parameters:	e		- (i) error code.
 * Returns:	Nothing.
 * Globals:	fsacl_errno	- (o) error code.
 * Exceptions:	None.
 * Remarks:	The error code is retrieved with fsacl_get_errno.
 */
void
fsacl_set_errno(const int e) {
  fsacl_errno = e;
}//fsacl_set_errno

/* Name:	fsacl_get_errno
 * Class:	None.
 * Purpose:	Returns the global error code for the library.
 * Parameters:	None.
 * Returns:	The error code.
 * Globals:	fsacl_errno	- (i) error code.
 * Exceptions:	None.
 * Remarks:	Since the library can also be used with languages other
 *		than C++, we cannot use exceptions externally.
 */
int
fsacl_get_errno(void) {
  return fsacl_errno;
}

/***	EOF of fsacl.cc	***/
