/***	fsacl.h	***/

/*	Copyright (C) Jan Daciuk, 2016	*/


typedef long int stateno_t;
typedef long int stateaddr_t;

// Error codes
#define	FSACL_OK		0 // no error
#define FSACL_NumberBigger	2 // serialized number (entry number) is bigger
                                  // then the number of bytes allocated for it
#define FSACL_InvalidTransNo	3 // Invalid transition number to be deleted
#define FSACL_NoMemory  	4 // Not enough memory for the program
#define FSACL_IllegalState  	5 // Invalid state number to be deleted
#define FSACL_InvSepAn  	6 // at least one entry has no annotations
                                  // or invalid annotation separator
#define FSACL_InvInpFile	7 // Could not open input file
#define FSACL_InvOutFile	8 // Could not create output file
#define FSACL_OtherErr	      100 // Error caught in STL and other libraries


// Options for building an automaton (used as bit masks)
#define		SORT_ON_FREQ	1
#define		WITH_NUMBERS	2
#define		OPTIMIZE	4
#define		MAKE_INDEX	8
#define		WEIGHTED	16
#define		GENERALIZE	64
#define		PROGRESS	128
#define		DESCENDING	256


#ifdef __cplusplus
extern "C" {
#endif

/* Name:	set_option
 * Class:	None.
 * Purpose:	Sets an option for building an automaton.
 * Parameters:	options		- (i/o) option variable to be used;
 *		option		- (i) options to be set.
 * Returns:	Nothing.
 * Globals:	None.
 * Exceptions:	None.
 * Remarks:	Options are written as 1-bit flags.
 */
inline void
set_option(unsigned int &options, unsigned int option) {
  options |= option;
};//set_option

/* Name:	write_sorted_construction_file
 * Class:	automaton
 * Purpose:	Constructs an automaton from sorted data, and writes it
 *		to a file.
 * Parameters:	input		- (i) input file name
 *					or empty for standard input;
 *		output		- (i) output file name
 *					or empty for standard output;
 *		options		- (i) options for building the automaton;
 *		annot_sep	- (i) annotation separator;
 *		filler		- (i) filler character;
 *		version		- (i) representation format version.
 * Returns:	Nothing.
 * Globals:	fsacl_errno	- (o) error code (0 is OK).
 * Exceptions:	None.
 * Remarks:	Invoke fsacl_get_errno right afterwards.
 */
void
write_sorted_construction_file(const char *input, const char *output,
			       const unsigned int options, const char annot_sep,
			       const char filler,
			       const unsigned int version);

/* Name:	write_unsorted_construction_file
 * Class:	None.
 * Purpose:	Constructs an automaton from unsorted data, and writes it
 *		to a file.
 * Parameters:	input		- (i) input file name
 *					or empty for standard input;
 *		output		- (i) output file name
 *					or empty for standard output;
 *		options		- (i) options for building the automaton;
 *		annot_sep	- (i) annotation separator;
 *		filler		- (i) filler character;
 *		version		- (i) representation format version.
 * Returns:	Nothing.
 * Globals:	fsacl_errno	- (o) error code (0 is OK).
 * Exceptions:	None.
 * Remarks:	Invoke fsacl_get_errno right afterwards.
 */
void
write_unsorted_construction_file(const char *input, const char *output,
				 const unsigned int options,
				 const char annot_sep,
				 const char filler,
				 const unsigned int version);


/* Name:	fsacl_set_errno
 * Class:	None.
 * Purpose:	Sets an error code.
 * Parameters:	e		- (i) error code.
 * Returns:	Nothing.
 * Globals:	fsacl_errno	- (o) error code.
 * Exceptions:	None.
 * Remarks:	The error code is retrieved with fsacl_get_errno.
 */
void
fsacl_set_errno(const int e);

/* Name:	fsacl_get_errno
 * Class:	None.
 * Purpose:	Returns the global error code for the library.
 * Parameters:	None.
 * Returns:	The error code.
 * Globals:	fsacl_errno	- (i) error code.
 * Exceptions:	None.
 * Remarks:	Since the library can also be used with languages other
 *		than C++, we cannot use exceptions externally.
 */
int
fsacl_get_errno(void);

#ifdef __cplusplus
}
#endif

/***	EOF fsacl.h	***/
