/***	nfsa_ubuild.cc	***/

/* Copyright (C) Jan Daciuk, 2016 */

#include	<string.h>
#include	<stdlib.h>
#include	<iomanip>
#include	<iostream>
#include	<fstream>
#include	"fsacl.h"

/*

  This program constructs a finite-state automaton from an unsorted list
  of words. It mimics behavior of fsa_ubuild from my fsa package.

*/

const char *version = "version 0.2, 6 Sept. 2016";

/* Name:	usage
 * Class:	None.
 * Purpose:	Prints program synopsis.
 * Parameters:	prog_name	- (i) program name.
 * Returns:	Nothing.
 * Remarks:	None.
 */
void
usage(const char *prog_name)
{
  std::cerr << prog_name
       << " builds an automaton. Synopsis:\n"
       << prog_name << " [options]\nOptions:\n-O\toptimization\n"
       << "-i input_file\t input file name (or else use stdin)\n"
       << "-o output_file\t output file name (or else use stdout)\n"
       << "-p\tshow progress\n"
       << "-v\tversion details\n"
       << "-A as\tannotation separator character (use with -X)\n"
       << "\t\t[default:+]\n"
       << "-D\tdescending sort (use with -S)\n"
       << "-F fc\tfiller character (use with -X)\n"
       << "\t\t[default:_]\n"
       << "-G\tgeneralize guessing automata\n"
       << "-N\tnumber entries (perfect hashing)\n"
       << "-P\tprefix mode (for guessing)\n"
       << "-S\tsort transitions on frequency of labels\n"
       << "-V format_number\trepresentation format version number\n"
       << "\t\t[default: 5]\n"
       << "-W\tweight arcs (for probabilities in guessing)\n"
       << "-X\tmake index a tergo (for guessing)\n"
       << "Example:\nsort -u word_list | "
       << prog_name << " -O > dict1.fsa\n";
}//usage

/* Name:	main
 * Class:	None.
 * Purpose:	Launches the program.
 * Parameters:	argc	- (i) number of program arguments;
 *		argv	- (i) program arguments.
 * Returns:	0	- if OK;
 *		1	- if invalid options;
 *		2	- if automaton could not be built;
 *		3	- if automaton could not be written to a file;
 *		4	- if there is not enough memory to build the automaton.
 * Globals:	version	- (i) version of the program.
 * Exceptions:	
 * Remarks:	None.
 */
int
main(const int argc, const char *argv[])
{
  unsigned int options = 0;
  unsigned int format_version = 5;
  const char *input_file_name = NULL;
  const char *output_file_name = NULL;
  char ANNOT_SEP = '+';
  char	FILLER = '_';

  if (argc >= 2) {
    for (int i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-O") == 0) {
	set_option(options, OPTIMIZE);
      }
      else if (strcmp(argv[i], "-X") == 0) {
	set_option(options, MAKE_INDEX);
      }
      else if (strcmp(argv[i], "-A") == 0) {
	if (++i < argc)
	  ANNOT_SEP = *argv[i];
	else {
	  std::cerr << "Invalid annotation separator\n";
	  usage(argv[0]);
	  return 1;
	}
      }
      else if (strcmp(argv[i], "-G") == 0) {
	set_option(options, GENERALIZE);
      }
      else if (strcmp(argv[i], "-p") == 0) {
	set_option(options, PROGRESS);
      }
      else if (strcmp(argv[i], "-i") == 0) {
	if (input_file_name) {
	  std::cerr << argv[0] << ": Multiple input files not supported"
		    << std::endl
		    << "Ignoring earlier inputs" << std::endl;
	}
	if (++i < argc)
	  input_file_name = argv[i];
	else {
	  std::cerr << argv[0] << ": -i without file name" << std::endl;
	  usage(argv[0]);
	  return 1;
	}
      }
      else if (strcmp(argv[i], "-o") == 0) {
	if (output_file_name) {
	  std::cerr << argv[0] << ": Multiple output files not supported"
		    << std::endl
		    << "Ignoring earlier output" << std::endl;
	}
	if (++i < argc)
	  output_file_name = argv[i];
	else {
	  std::cerr << argv[0] << ": -o without file name" << std::endl;
	  usage(argv[0]);
	  return 1;
	}
      }
      else if (strcmp(argv[i], "-F") == 0) {
	if (++i < argc) {
	  FILLER = argv[0][0];
	}
	else {
	  std::cerr << argv[0] << ": -F without a character" << std::endl;
	  usage(argv[0]);
	  return 1;
	}
      }
      else if (strcmp(argv[i], "-N") == 0) {
	set_option(options, WITH_NUMBERS);
      }
      else if (strcmp(argv[i], "-V") == 0) {
	if (++i < argc) {
	  format_version = atoi(argv[i]);
	}
	else {
	  std::cerr << argv[0] << ": -V without version number" << std::endl;
	  usage(argv[0]);
	  return 1;
	}
      }
      else if (strcmp(argv[i], "-W") == 0) {
	set_option(options, WEIGHTED);
      }
      else if (strcmp(argv[i], "-P") == 0) {
	set_option(options, PREFIX_MODE);
      }
      else if (strcmp(argv[i], "-S") == 0) {
	set_option(options, SORT_ON_FREQ);
      }
      else if (strcmp(argv[i], "-D") == 0) {
	set_option(options, DESCENDING);
      }
      else if (strcmp(argv[i], "-v") == 0) {
	std::cerr << "nfsa_build " << version << std::endl;
	return 0;
      }
      else {
	usage(argv[0]);
	return 1;
      }
    }//for
  }//if

  if ((options & (WITH_NUMBERS | OPTIMIZE)) ==
      (WITH_NUMBERS | OPTIMIZE)) {
    std::cerr << "-N and -O cannot be specified together. Turning -O off"
	      << std::endl;
    options &= ~OPTIMIZE;
  }
  //set_new_handler(&not_enough_memory);

  write_unsorted_construction_file(input_file_name, output_file_name, options,
				   ANNOT_SEP, FILLER, format_version);
  int e = fsacl_get_errno();
  if (e) {
    switch (e) {
    case FSACL_NumberBigger:
      std::cerr << "nfsa_ubuild: internal error (bug):" << std::endl
		<< "Trying to write a big number on insufficient "
		<< "number of bytes" << std::endl;
      break;

    case FSACL_InvalidTransNo:
      std::cerr << "nfsa_ubuild internal error (bug):" << std::endl
		<< "Invalid transition number when deleting a transition"
		<< std::endl;
      break;

    case FSACL_NoMemory:
      std::cerr << "nfsa_ubuild error: insufficient memory" << std::endl;
      break;

    case FSACL_IllegalState:
      std::cerr << "nfsa_ubuild internal error (bug):" << std::endl
		<< "Invalid state number when trying to delete a state"
		<< std::endl;

    case FSACL_InvSepAn:
      std::cerr << "nfsa_ubuild: error in input data." << std::endl
		<< "Invalid or no annotation separator" << std::endl;
      break;

    case FSACL_InvInpFile:
      std::cerr << "nfsa_ubuild: could not open input file `"
		<< input_file_name << "'" << std::endl;
      break;

    case FSACL_InvOutFile:
      std::cerr << "nfsa_ubuild: could not open output file `"
		<< output_file_name << "'" << std::endl;
      break;

    case FSACL_OtherErr:
      std::cerr << "nfsa_ubuild: unspecified error detected in STL or other "
		<< " libraries" << std::endl;
      break;
    }
  }
  return e;
}//main

/*** EOF nfsa_ubuild.cc	***/
